package me.parsell.dmblock.Common;

import java.nio.FloatBuffer;
import java.util.Map;
import java.util.Set;

import org.lwjgl.system.MemoryUtil;

import net.minecraft.block.Block;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Matrix3f;
import net.minecraft.util.math.Vec3i;
import net.minecraft.world.World;

public abstract class MultiBlockStructure {
    public class MultiBlockInstance{
        public Vec3i mutliBlockEntry;
        public RotationDirection rotation;

        public MultiBlockInstance(Vec3i entry, RotationDirection rotation){
            mutliBlockEntry = entry;
            this.rotation = rotation;
        }
    }

    protected enum RotationDirection{
        x0, x90, x180, x270,
        y0, y90, y180, y270,
        z0, z90, z180, z270
    }

    protected final RotationDirection[] ALL_ROTATION_DIRECTIONS = {RotationDirection.x0, RotationDirection.x90, RotationDirection.x180, RotationDirection.x270,
                                                                   RotationDirection.y0, RotationDirection.y90, RotationDirection.y180, RotationDirection.y270,
                                                                   RotationDirection.z0, RotationDirection.z90, RotationDirection.z180, RotationDirection.z270};
    protected final RotationDirection[] NO_ROTATION_DIRECTIONS = {RotationDirection.x0, RotationDirection.y0, RotationDirection.z0};
    protected final RotationDirection[] ALL_X_ROTATION_DIRECTONS = {RotationDirection.x0, RotationDirection.x90, RotationDirection.x180, RotationDirection.x270};
    protected final RotationDirection[] ALL_Y_ROTATION_DIRECTONS = {RotationDirection.y0, RotationDirection.y90, RotationDirection.y180, RotationDirection.y270};
    protected final RotationDirection[] ALL_Z_ROTATION_DIRECTONS = {RotationDirection.z0, RotationDirection.z90, RotationDirection.z180, RotationDirection.z270};

    public abstract Map<Vec3i, Block> getStructure();
    public abstract Set<RotationDirection> getValidRotationDirections();

    public MultiBlockInstance verifyStructure(World world, BlockPos activation){
        Map<Vec3i, Block> structureTemplate = getStructure();
        for(Map.Entry<Vec3i, Block> startEntry: structureTemplate.entrySet()){ // Iterate through all blocks in the structure so we can determine where on the structure the player activated
            if(world.getBlockState(activation).isOf(startEntry.getValue())){ // Ensure that the block we're checking is the block the player activated on
                // Block in template is same the player activated on, this is our entry block for this formation check
                Vec3i startPos = startEntry.getKey(); // Get the relative position in the structure
                for(RotationDirection rotation : getValidRotationDirections()){ // Iterate through all possible rotations of the structure as well
                    boolean wasRotationVerified = true; // To determine if the structure can be formed given this entry block and this rotation
                    for(Map.Entry<Vec3i, Block> entry: structureTemplate.entrySet()){ // Iterate through all remaining blocks in the structure
                        if(entry.getKey().equals(startPos)) // if its one we already checked, skip
                            continue;

                        Vec3i offset = applyRotationMatrix(startPos, entry.getKey(), rotation); // Modify the relative position we are checking by the entry position's location
                        BlockPos checkPosition = activation.add(offset); // Offset the actual blockpos with the relative one, corrected for the entry blockpos
                        if(!(world.getBlockState(checkPosition).isOf(entry.getValue()))){ // Check block against structure block
                            wasRotationVerified = false; // If the block was not the block in the structure, then this rotation was impossible and continue to check other rotations
                            break;
                        }
                    }
                    if (wasRotationVerified){ // If one of the rotations passed, this means formation is possible
                        return new MultiBlockInstance(startPos, rotation);
                    }
                }
            }
        }
        return null; // No formation was possible
    }
    
    // This implements a 3d rotation matrix to modify our relative blockpos for the structure
    public static Vec3i applyRotationMatrix(Vec3i origin, Vec3i destination, RotationDirection rotation){
        FloatBuffer marcusFloats = MemoryUtil.memAllocFloat(9);
        rotationToMatrix(rotation).writeColumnMajor(marcusFloats);
        Vec3i rawOffset = destination.subtract(origin); // It's fucking raw
        float x = (float)rawOffset.getX();
        float y = (float)rawOffset.getY();
        float z = (float)rawOffset.getZ();
        Vec3i finalOffset = new Vec3i(x * marcusFloats.get(0) + y * marcusFloats.get(1) + z * marcusFloats.get(2), 
                                      x * marcusFloats.get(3) + y * marcusFloats.get(4) + z * marcusFloats.get(5), 
                                      x * marcusFloats.get(6) + y * marcusFloats.get(7) + z * marcusFloats.get(8)); //Cooked up with some crack
        return finalOffset;
    }

    // This allows us to take human-readable rotations and convert it to the correct matrix
    // This cannot be a HashMap or the likes due to Matrix3f not allowing intilization with data in a simple way, thanks mojang
    public static Matrix3f rotationToMatrix(RotationDirection rotation){
        Matrix3f rotMat = new Matrix3f();
        switch(rotation) {
            case x0:
            case y0:
            case z0:
                rotMat.set(0, 0, 1);
                rotMat.set(0, 1, 0);
                rotMat.set(0, 2, 0);
                rotMat.set(1, 0, 0);
                rotMat.set(1, 1, 1);
                rotMat.set(1, 2, 0);
                rotMat.set(2, 0, 0);
                rotMat.set(2, 1, 0);
                rotMat.set(2, 2, 1);
                return rotMat;
            case x90:
                rotMat.set(0, 0, 1);
                rotMat.set(0, 1, 0);
                rotMat.set(0, 2, 0);
                rotMat.set(1, 0, 0);
                rotMat.set(1, 1, 0);
                rotMat.set(1, 2, -1);
                rotMat.set(2, 0, 0);
                rotMat.set(2, 1, 1);
                rotMat.set(2, 2, 0);
                return rotMat;
            case x180:
                rotMat.set(0, 0, 1);
                rotMat.set(0, 1, 0);
                rotMat.set(0, 2, 0);
                rotMat.set(1, 0, 0);
                rotMat.set(1, 1, -1);
                rotMat.set(1, 2, 0);
                rotMat.set(2, 0, 0);
                rotMat.set(2, 1, 0);
                rotMat.set(2, 2, -1);
                return rotMat;
            case x270:
                rotMat.set(0, 0, 1);
                rotMat.set(0, 1, 0);
                rotMat.set(0, 2, 0);
                rotMat.set(1, 0, 0);
                rotMat.set(1, 1, 0);
                rotMat.set(1, 2, 1);
                rotMat.set(2, 0, 0);
                rotMat.set(2, 1, -1);
                rotMat.set(2, 2, 0);
                return rotMat;
            case y90:
                rotMat.set(0, 0, 0);
                rotMat.set(0, 1, 0);
                rotMat.set(0, 2, 1);
                rotMat.set(1, 0, 0);
                rotMat.set(1, 1, 1);
                rotMat.set(1, 2, 0);
                rotMat.set(2, 0, -1);
                rotMat.set(2, 1, 0);
                rotMat.set(2, 2, 0);
                return rotMat;
            case y180:
                rotMat.set(0, 0, -1);
                rotMat.set(0, 1, 0);
                rotMat.set(0, 2, 0);
                rotMat.set(1, 0, 0);
                rotMat.set(1, 1, 1);
                rotMat.set(1, 2, 0);
                rotMat.set(2, 0, 0);
                rotMat.set(2, 1, 0);
                rotMat.set(2, 2, -1);
                return rotMat;
            case y270:
                rotMat.set(0, 0, -1);
                rotMat.set(0, 1, 0);
                rotMat.set(0, 2, 0);
                rotMat.set(1, 0, 0);
                rotMat.set(1, 1, 1);
                rotMat.set(1, 2, 0);
                rotMat.set(2, 0, 1);
                rotMat.set(2, 1, 0);
                rotMat.set(2, 2, 0);
                return rotMat;
            case z90:
                rotMat.set(0, 0, 0);
                rotMat.set(0, 1, -1);
                rotMat.set(0, 2, 0);
                rotMat.set(1, 0, 1);
                rotMat.set(1, 1, 0);
                rotMat.set(1, 2, 0);
                rotMat.set(2, 0, 0);
                rotMat.set(2, 1, 0);
                rotMat.set(2, 2, 1);
                return rotMat;
            case z180:;
                rotMat.set(0, 0, -1);
                rotMat.set(0, 1, 0);
                rotMat.set(0, 2, 0);
                rotMat.set(1, 0, 0);
                rotMat.set(1, 1, -1);
                rotMat.set(1, 2, 0);
                rotMat.set(2, 0, 0);
                rotMat.set(2, 1, 0);
                rotMat.set(2, 2, 1);
                return rotMat;
            case z270:
                rotMat.set(0, 0, 0);
                rotMat.set(0, 1, 1);
                rotMat.set(0, 2, 0);
                rotMat.set(1, 0, -1);
                rotMat.set(1, 1, 0);
                rotMat.set(1, 2, 0);
                rotMat.set(2, 0, 0);
                rotMat.set(2, 1, 0);
                rotMat.set(2, 2, 1);
                return rotMat;
            default:
                return null; // Something fucked up bad here, blame someone
        }
    }
}